/*-------------------------------------------------------------------------+
| Copyright 2013 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.service;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstChildWithType;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.SimpleDateFormat;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.project.model.FileProject;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.ext.quality.service.IModelQualityProvider.IVisitorModelQualityProvider;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IMigrationService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;

/**
 * This class implements the {@link IMigrationService} interface.
 * 
 * @author blaschke
 */
public class ModelQualityService extends EObjectAwareServiceBase<IModelQualityProvider<?>>
		implements IIntrospectiveKernelService, IModelQualityService {

	/** The singleton instance. */
	private static final ModelQualityService INSTANCE = new ModelQualityService();

	/** Returns singleton instance of the service. */
	public static ModelQualityService getInstance() {
		return INSTANCE;
	}

	/** The connector extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.ext.quality.metricService";

	/** The connector configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "metricService";

	/** The connector attribute name. */
	private static final String ATTRIBUTE_NAME = "metricService";

	/** Static Map containing the last result of the metric extraction */
	public static Map<IProjectRootElement, IModelQualityProvider<? extends EObject>> last_metrics;

	/** */
	@Override
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** Registers the migration provider with the service. */
	@Override
	public void registerMetricProvider(IModelQualityProvider<?> provider,
			Class<?> modelElementClass) {
		addHandler(modelElementClass, provider);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service allows the tracking of metrical development in modeling." +
				"\n the service outputs the current metric analysis of the model and stores" +
				"\n the data within logs for later history analysis." +
				"\n\nThe service extension point is '" + EXTENSION_POINT_NAME + "'.";
	}

	/** Get all suitable {@link IModelQualityProvider} for the given input. */
	protected List<IModelQualityProvider<?>> getAllMetricProviders(EObject input) {
		List<IModelQualityProvider<?>> providers = new ArrayList<>();
		for(Entry<Class<?>, List<IModelQualityProvider<?>>> migEntry : handlerMap.entrySet()) {
			if(getFirstChildWithType(input, migEntry.getKey()) != null) {
				providers.addAll(migEntry.getValue());
			}
		}
		System.out.println("These are the providers:" + providers);
		return providers;
	}

	/**
	 * @param fp
	 *            the file project on which the analysis should be performed
	 * @return result resulting metrics
	 */
	public Map<IProjectRootElement, IModelQualityProvider<? extends EObject>>
			performMetricAnalysis(FileProject fp) {
		Map<IProjectRootElement, IModelQualityProvider<? extends EObject>> result = new HashMap<>();
		for(IProjectRootElement rootElement : fp.getRootElements()) {
			if(rootElement instanceof IHierarchicElement) {

				// TODO check get(0)
				IHierarchicElement firstNode =
						((IHierarchicElement)rootElement).getContainedElements().get(0);

				HierarchicElementSizeProvider hes = new HierarchicElementSizeProvider();
				hes.apply(firstNode);
				System.out.println("checked in hierarchic metrics for: " + rootElement.getName());
				double commentable = hes.getStoredMetrics().getOrDefault(
						HierarchicElementSizeProvider.NUMBER_OF_TOTAL_COMMENTABLE_ELEMENTS, 0.0);
				double commented = hes.getStoredMetrics().getOrDefault(
						HierarchicElementSizeProvider.NUMBER_OF_TOTAL_COMMENTED_ELEMENTS, 0.0);
				System.out.println("Commented to total: " + commented + ", " + commentable + ", " +
						commented / commentable);
				result.put(rootElement, hes);
			}
		}

		// List<IModelQualityProvider<?>> handlers = getRegisteredHandlers(modelElement.getClass());
		// Map<String, Double> qualityData = new HashMap<String, Double>();

		// getRegisteredHandlers(IHierarchicElement.class);
		// List<IModelQualityProvider<?>> hierarchicHandlers2 = getAllMetricProviders(fp);
		// if(handlers == null) {
		// return null;
		// }

		// for(IModelQualityProvider<?> qualityProv : handlers) {
		// List<? extends IModelQualityProvider<? extends EObject>> collectedProviders =
		// qualityProv.apply(modelElement);
		// if(metrics != null) {
		// metricsList.addAll(metrics);
		// }
		// }

		last_metrics = result;
		metricExtractionToCSV(result);
		return result;
	}

	/** */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** */
	@Override
	protected String getHandlerClassAttribute() {
		return ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Metric Service";
	}

	/** {@inheritDoc} */
	@Override
	public List<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}

	/**
	 * Method to write the data into a csv
	 * 
	 * @param data
	 *            to convert into csv
	 * 
	 */
	public void metricExtractionToCSV(
			Map<IProjectRootElement, IModelQualityProvider<? extends EObject>> data) {

		// check if there is any data to export
		if(data == null || data.isEmpty()) {
			return;
		}
		
		// path to csv file
		Path path = Paths.get("data_ultra_cool.csv");

		List<String> allKeys = new ArrayList<>();
		boolean createNewIndex = true;
		
		if(Files.exists(path)) {
			// Read first line and try to extract the keys from the already existing file
			try(BufferedReader reader = new BufferedReader(new FileReader(path.toFile()))) {
				String line = reader.readLine();
				if(line != null) {
					allKeys = Arrays.asList(line.split(","));
					createNewIndex = false;
				}
			} catch(IOException e) {
				e.printStackTrace();
			}
		}

		try(var writer = new BufferedWriter(new FileWriter(path.toFile(), true))) {
			
			if(createNewIndex) {
				// Create new index and write it into the first line of the file
				var anyProvider = data.values().iterator().next();
				List<String> keys = anyProvider.metricKeys();
				allKeys.add("timestamp");
				allKeys.add("root");
				allKeys.add("node");
				allKeys.add("parent");
				allKeys.addAll(keys);

				// Write first line
				writer.write(String.join(",", allKeys));
				writer.write("\n");
			}
			// Remove first 4 keys as they are not provided by the provider
			List<String> valueKeys = allKeys.subList(4, allKeys.size());

			LocalDateTime dateTime = LocalDateTime.now();
			DateTimeFormatter formatter = DateTimeFormatter.ISO_LOCAL_DATE_TIME;
			String formattedDateTime = dateTime.format(formatter);

			data.forEach((rootProject, rootProvider) -> {
				// iterate over all project roots
				String rootName = rootProject.getName();

				rootProvider.traverse(new IVisitorModelQualityProvider() {

					@Override
					public void accept(HierarchicElementSizeProvider provider) {
						
						var metrics = provider.getStoredMetrics();
						// These 4 data entries are just concatenated in front
						var startStream = Stream.of(formattedDateTime, rootName, "null", "null");
						// Collect all values from provider in the correct order and combine 
						var values = Stream
								.concat(startStream,
										valueKeys.stream().map(metrics::get).map(String::valueOf))
								.collect(Collectors.joining(","));

						try {
							// write values 
							writer.write(values);
							writer.write("\n");
						} catch(IOException e) {
							e.printStackTrace();
						}
					}
				});
			});
		} catch(IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * 
	 * @author blaschke
	 */
	public static class MetricCSVWriter {
		/**  */
		private static final String FILE_PATH = "data_ultra_cool.csv";
		/**  */
		private BufferedWriter writer;

		/**
		 * @throws IOException
		 */
		public void openFile() throws IOException {
			writer = new BufferedWriter(new FileWriter(FILE_PATH, true));
		}

		/**
		 * @param data
		 */
		public void writeData(Map<String, Double> data) {
			try {
				String timestamp = getCurrentTimestamp();
				writer.write(timestamp + ": ");
				for(Map.Entry<String, Double> entry : data.entrySet()) {
					writer.write(entry.getKey() + ":" + entry.getValue() + ",");
					writer.newLine();
				}
				writer.flush();
			} catch(IOException e) {
				e.printStackTrace();
			}
		}

		/**
		 * @throws IOException
		 */
		public void closeFile() throws IOException {
			if(writer != null) {
				writer.close();
			}
		}

		/** */
		private String getCurrentTimestamp() {
			SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
			Date now = new Date();
			return dateFormat.format(now);
		}
	}
}
